/*
** 1998-09-12 -	A Move As command. Very useful.
** 1999-03-06 -	Adapted for the new selection/generic/dirrow representations.
*/

#include "gentoo.h"
#include "errors.h"
#include "fileutil.h"
#include "overwrite.h"
#include "progress.h"
#include "dirpane.h"
#include "strutil.h"

#include "cmd_move.h"
#include "cmd_generic.h"

#include "cmd_moveas.h"

#define	CMD_ID	"move as"

/* ----------------------------------------------------------------------------------------- */

typedef struct {
	GtkWidget	*vbox;
	GtkWidget	*label;
	GtkWidget	*entry;
	MainInfo	*min;
	gint		ovw_open;
} MvaInfo;

/* ----------------------------------------------------------------------------------------- */

/* 1998-09-12 -	Update body of copy as GUI. */
static void mva_body(MainInfo *min, DirPane *src, DirRow *row, gpointer user)
{
	gchar	buf[FILENAME_MAX + 32];
	MvaInfo	*mva = user;

	g_snprintf(buf, sizeof buf, _("Enter Name to Move \"%s\" As"), DP_ROW_NAME(row));
	gtk_label_set_text(GTK_LABEL(mva->label), buf);
	gtk_entry_set_text(GTK_ENTRY(mva->entry), DP_ROW_NAME(row));
	gtk_entry_select_region(GTK_ENTRY(mva->entry), 0, -1);
	gtk_widget_grab_focus(mva->entry);

	if(mva->ovw_open == FALSE)
	{
		ovw_overwrite_begin(mva->min, _("\"%s\" Already Exists - Continue With Move?"), 0U);
		mva->ovw_open = TRUE;
	}
}

static gint mva_action(MainInfo *min, DirPane *src, DirPane *dst, DirRow *row, gpointer user)
{
	gchar		dest[PATH_MAX], *text;
	MvaInfo		*mva = user;
	gboolean	ret = FALSE;

	if(!fut_cd(src->dir.path, NULL, 0))
		return 0;

	text = gtk_entry_get_text(GTK_ENTRY(mva->entry));
	if(text && *text)
	{
		if(strchr(text, G_DIR_SEPARATOR) == NULL)
			g_snprintf(dest, sizeof dest, "%s/%s", dst->dir.path, text);
		else
			stu_strncpy(dest, text, sizeof dest);
		pgs_progress_begin(min, _("Moving As..."), PFLG_COUNT_RECURSIVE | PFLG_ITEM_VISIBLE | PFLG_BYTE_VISIBLE);
		if(S_ISDIR(DP_ROW_LSTAT(row).st_mode))
			ret = move_dir(min, DP_ROW_NAME(row), dest);
		else
			ret = move_file(min, DP_ROW_NAME(row), dest, &DP_ROW_STAT(row));
		pgs_progress_end(min);
	}
	if(ret)
		dp_unselect(src, DP_ROW_INDEX(src, row));

	return ret && errno == 0;
}

static void mva_free(gpointer user)
{
	MvaInfo	*mva = user;

	if(mva->ovw_open)
		ovw_overwrite_end(mva->min);
}

/* 1998-09-12 -	The move as generic command function. */
gint cmd_moveas(MainInfo *min, DirPane *src, DirPane *dst, const CmdArg *ca)
{
	static MvaInfo	mva;

	mva.min	     = min;
	mva.ovw_open = FALSE;

	mva.vbox = gtk_vbox_new(FALSE, 0);
	mva.label = gtk_label_new(_("Move As"));
	mva.entry = gtk_entry_new_with_max_length(FILENAME_MAX - 1);
	gtk_box_pack_start(GTK_BOX(mva.vbox), mva.label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(mva.vbox), mva.entry, FALSE, FALSE, 0);

	return cmd_generic(min, _("Move As"), CGF_NOALL | CGF_SRC, mva_body, mva_action, mva_free, &mva);
}
