/*
** 1998-05-23 -	A big fat error handling module. There are *so* many possible errors,
**		I think we really must get a grip on that situation. This is it.
**		As far as grips go, I consider this to be pretty loose. :(
*/

#include "gentoo.h"

#include <ctype.h>
#include <stdarg.h>
#include <string.h>

#include "dirpane.h"
#include "dialog.h"
#include "errors.h"

/* ----------------------------------------------------------------------------------------- */

static gchar	error_desc[1024] = "";

/* ----------------------------------------------------------------------------------------- */

void err_clear(MainInfo *min)
{
	error_desc[0] = '\0';
	errno = 0;
}

/* 1999-01-21 -	Format an arbitrary string and print it on the status (error) line.
** UGLY   UGLY	Uses a sneaky little loop to ensure that the text is displayed *NOW*.
*/
void err_printf(MainInfo *min, const gchar *fmt, ...)
{
	va_list	al;

	va_start(al, fmt);
	vsnprintf(error_desc, sizeof error_desc, fmt, al);
	va_end(al);
	err_show(min);
	while(gtk_events_pending())		/* This might qualify as a kludge. */
		gtk_main_iteration();
}

/* 1998-05-30 -	Added special treatment of code == -1; this means that there is no
**		errno-compatible error information available, and the entire thing
**		is then suppressed.
*/
gint err_set(MainInfo *min, gint code, const gchar *source, const gchar *obj)
{
	gint	old_errno = errno;

	if(source != NULL)
	{
		if(code >= 0)
		{
			gchar	*desc;

			desc = g_strerror(code);
			if(obj != NULL)
				g_snprintf(error_desc, sizeof error_desc, _("Couldn't %s \"%s\": %s (code %d)"), source, obj, desc, code);
			else
				g_snprintf(error_desc, sizeof error_desc, _("Couldn't %s \"%s\" (code %d)"), source, obj, code);
		}
		else
		{
			if(obj != NULL)
				g_snprintf(error_desc, sizeof error_desc, _("Couldn't %s \"%s\""), source, obj);
			else
				g_snprintf(error_desc, sizeof error_desc, _("Couldn't %s"), source);
		}
	}
	errno = old_errno;
	return 1;
}

void err_show(MainInfo *min)
{
	if(error_desc[0] != '\0')
	{
		gtk_label_set_text(GTK_LABEL(min->gui->top), error_desc);
		while(gtk_events_pending())
			gtk_main_iteration();
		if(min->cfg.errors.dialog)
			dlg_dialog_async_new_simple(error_desc, "Error", "OK", NULL, NULL);
		if(min->cfg.errors.beep)
			gdk_beep();
	}
	else
		dp_show_stats(min->gui->cur_pane);
}
